//Dust Signs Calendar (dscal.pas)
//  Kleine, flexible Kalenderkomponente, basierend auf Canvas-
//  routinen von TCustomPanel bzw. TWinControl (vgl. unten)
//
//  Autor:    Andreas Unterweger, andi_uw@gmx.at
//  Homepage: http://www.dustsigns.de.vu/
//
//  Ausfhrliche Beschreibungen zu den Funktionen und Eigenschaften
//  finden Sie in der beigelegten Hilfedatei, Infos in der Readme.
//
//  (c) by Dust Signs Andreas Unterweger 2005
//
//  Lizenz:
//  Diese Datei darf weitergegeben, in kostenlos verfgbaren Programmen
//  verwendet und auch verndert werden, solange keiner der hier stehenden
//  Kommentare gelscht oder modifziert wird. Das Hinzufgen von eigenen
//  Kommentaren ist selbstverstndlich gestattet, solange diese die
//  Originalkommentare nicht in irgendeiner Form sinnentfremden.
//  Bei eventuellen Lizenzunklarheiten wenden Sie sich bitte an den
//  Autor (o.a. Email-Adresse oder Kontaktadresse auf der o.a. Homepage)
//
//  Anmerkung: zur bersetzung der Feiertagsnamen und anderen Wrtern
//  wurde der Service http://dict.leo.org/ verwendet.
//
//  Zuletzt bearbeitet: 09.05.2005, 20:23

unit dscal;

//{$DEFINE WINCONTROL}
//Wenn aktiviert, wird TDSCalendar direkt von TWinControl und
//nicht von TCustomPanel abgeleitet. Dazu ist ein eigenes Canvas
//notwendig, das in der private-Variable FCanvas gespeichert wird.
//Wichtiger Hinweis: Die Ableitung von TWinControl ist durch das
//Neuzeichnen der Button in puncto Performance bis zu fnf Mal
//langsamer als die Ableitung von TCustomPanel und sollte daher
//nur dann verwendet werden, wenn die Geschwindigkeit eine eher
//untergeordnetere Rolle spielt.

interface

uses
  {$IFNDEF WINCONTROL}
    ExtCtrls, //fr TCustomPanel
  {$ENDIF}
  Windows, Controls, SysUtils, Graphics, Classes, DateUtils, StdCtrls, Messages;

//Konstanten, die bereits im Interfaceteil bentigt werden
const
  //Fr "keine Auswahl"
  NoSelStart = 127;
  NoSelEnd = 127;

//Typendeklarationen
type
  //Tag-/Monat-Record
  TDayMonthRecord = record
                      Day, Month: Byte;
                      end;

  //Feiertagsarray (fr alle zwlf Monate ( 31 Tage); fixe Feiertage)
  TFixedHolidays = Array [1..12] of set of 1..31;

  //Spezielle Flags (z.B. Feiertag)
  TDateFlag = (dfNone, dfOutOfMonth, dfHoliday);

  //Koordinaten eines gezeichneten Tages
  TDayCoord = record
                Day: ShortInt; //Ausgewhlter Tag
                Date: TDateTime; //Ausgewhltes Datum
                SpecialFlag: TDateFlag;
                Coords: TRect; //Koordinaten des Textrechtecks
                end;

  //Koordinatenarray (Positionen der Tage fr Hints und Maus-Events)
  TDayCoords = Array [1..42] of TDayCoord; //Maximal 42 Elemente, da nicht mehr auf die Zeichenflche passen

  //Typ fr Zeilenindex
  TRowIndex = 1..6;

  //Ereignistypen-Deklarationen
  TDateChangeEvent = procedure(Sender: TObject; OldDate, NewDate: TDateTime; var Change: Boolean) of object;
  TSelectDateEvent = procedure(Sender: TObject; SelectedDay: ShortInt; XCoord, YCoord: Integer; SelectedDate: TDateTime; SpecialFlag: TDateFlag) of object;
  TRightClickDateEvent = procedure(Sender: TObject; XCoord, YCoord: Integer; DateAtCursor: TDateTime; SpecialFlag: TDateFlag) of object;
  TCustomDateHintEvent = procedure(Sender: TObject; Day: ShortInt; XCoord, YCoord: Integer; DateAtCursor: TDateTime; SpecialFlag: TDateFlag; var AHint: String) of object;

  //TDSCalendar
  {$IFDEF WINCONTROL}
    TDSCalendar = class (TWinControl)
  {$ELSE}
    TDSCalendar = class (TCustomPanel)
  {$ENDIF}  
  private
    //Canvas falls von TWinControl abgeleitet
    {$IFDEF WINCONTROL}
      FCanvas: TCanvas;
    {$ENDIF}

    //DoubleBuffer fr Zeichenflche
    DoubleBuffer: TBitmap;

    //Fixe Feiertage
    FFixedHolidays: TFixedHolidays;

    //Koordinaten fr Tage (fr OnMouseDown)
    FDayCoords: TDayCoords;
    CoordCount: Byte; //Anzahl gespeicherter Koordinaten

    //Fr SelectOnMouseMove
    FMouseIsDown: Boolean; //Fr SelectOnMouseMove: true zwischen MouseDown und MouseUp
    FMonthChanged: Boolean; //Wenn Monat gendert true um Auswahl rckzusetzen
    
    //Um Neuzeichnen beim Zerstren zu verhindern
    FIsDestroying: Boolean;

    //Buttons 'Vor' und 'Zurck'
    FNextMonthButton: TButton;
    FLastMonthButton: TButton;
    FNextYearButton: TButton;
    FLastYearButton: TButton;

    //Hintfenster
    FHintWnd: THintWindow;

    //Liste fr dynamische Feiertage
    FHolidayList: TList;

    //Zum Speichern der Monate als Day-Wert (kompatibel zu SelStartDay etc.)
    FMondaysDayArray: Array [TRowIndex] of ShortInt;

    //Fr SelStart bei MouseDown, wenn Datumsnderung unterbunden wird -> SelStartDay nicht setzen
    FNoSelStartAllowed: Boolean;

    //Koordinaten fr Wochen
    FWeekCoords: Array [TRowIndex] of TRect;

    //Fr Properties bentigte Variablen und Funktionen
    FDate: TDateTime;
    FOnChangeDate: TNotifyEvent;
    FOnChangeDateEx: TDateChangeEvent;
    FFont: TFont;
    FBorderColor: TColor;
    FBackgroundColor: TColor;
    FSundayColor: TColor;
    FWeekColor: TColor;
    FOutOfMonthColor: TColor;
    FSelectedDayBold: Boolean;
    FHeadlineStyle: TFontStyles;
    FTitleFont: TFont;
    FSelStartDay: ShortInt;
    FSelEndDay: ShortInt;
    FSelectionBackgroundColor: TColor;
    FAutoHeight: Boolean;
    FHolidayColor: TColor;
    FUseDoubleBuffer: Boolean;
    FOnSelectDate: TSelectDateEvent;
    FApplyOnSelect: Boolean;
    FMonthButtonsVisible: Boolean;
    FYearButtonsVisible: Boolean;
    FButtonSize: Integer;
    FHighlightDate: Boolean;
    FHeadlineOffset: Integer;
    FRowHeight: Integer;
    FReactOnMouseUp: Boolean;
    FReactOnMouseDown: Boolean;
    FOnRightClickDate: TRightClickDateEvent;
    FForceSixthWeekDraw: Boolean;
    FIgnoreHolidays: Boolean;
    FBorderWidth: TBorderWidth;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FButtonSpace: Integer;
    FAutoArrangeButtons: Boolean;
    FEnableHints: Boolean;
    FOnCustomDateHint: TCustomDateHintEvent;
    FOnBeforeMouseMove: TMouseMoveEvent;
    FOnBeforeMouseDown: TMouseEvent;
    FOnBeforeMouseUp: TMouseEvent;
    FOnAfterMouseMove: TMouseMoveEvent;
    FOnAfterMouseDown: TMouseEvent;
    FOnAfterMouseUp: TMouseEvent;
    FSelectOnMouseMove: Boolean;
    FOnSelectionChange: TNotifyEvent;
    FSaturdayColor: TColor;
    FAllowWeekSelection: Boolean;

    //Button-OnClick-Ereignisse
    procedure FLastMonthClick(Sender: TObject);
    procedure FNextMonthClick(Sender: TObject);
    procedure FLastYearClick(Sender: TObject);
    procedure FNextYearClick(Sender: TObject);

    //Ereignisse, die auf das ndern der Fonts reagieren
    procedure OnFontChange(Sender: TObject);
    procedure OnTitleFontChange(Sender: TObject);

    //Funktionen und Prozeduren fr Properties

    procedure SetDate(AValue: TDateTime);

    procedure SetOnChangeDate(AValue: TNotifyEvent);

    procedure SetOnChangeDateEx(AValue: TDateChangeEvent);

    procedure SetFont(AValue: TFont);

    procedure SetBorderColor(AValue: TColor);

    procedure SetBackgroundColor(AValue: TColor);

    procedure SetSundayColor(AValue: TColor);

    procedure SetWeekColor(AValue: TColor);

    procedure SetOutOfMonthColor(AValue: TColor);

    procedure SetSelectedDayBold(AValue: Boolean);

    procedure SetHeadlineStyle(AValue: TFontStyles);

    procedure SetTitleFont(AValue: TFont);

    procedure SetSelStartDay(AValue: ShortInt);

    procedure SetSelEndDay(AValue: ShortInt);

    procedure SetSelectionBackgroundColor(AValue: TColor);

    procedure SetAutoHeight(AValue: Boolean);

    procedure SetHolidayColor(AValue: TColor);

    procedure SetUseDoubleBuffer(AValue: Boolean);

    procedure SetOnSelectDate(AValue: TSelectDateEvent);

    procedure SetApplyOnSelect(AValue: Boolean);

    procedure SetMonthButtonsVisible(AValue: Boolean);

    procedure SetYearButtonsVisible(AValue: Boolean);

    procedure SetButtonSize(AValue: Integer);

    function GetLastMonthButtonHint: String;
    procedure SetLastMonthButtonHint(AValue: String);

    function GetLastYearButtonHint: String;
    procedure SetLastYearButtonHint(AValue: String);

    function GetNextMonthButtonHint: String;
    procedure SetNextMonthButtonHint(AValue: String);

    function GetNextYearButtonHint: String;
    procedure SetNextYearButtonHint(AValue: String);

    procedure SetHighlightDate(AValue: Boolean);

    procedure SetHeadlineOffset(AValue: Integer);

    procedure SetRowHeight(AValue: Integer);

    procedure SetReactOnMouseDown(AValue: Boolean);

    procedure SetReactOnMouseUp(AValue: Boolean);

    procedure SetRightClickDate(AValue: TRightClickDateEvent);

    procedure SetForceSixthWeekDraw(AValue: Boolean);

    procedure SetIgnoreHolidays(AValue: Boolean);

    procedure SetBorderWidth(AValue: TBorderWidth);

    procedure SetOnMouseEnter(AValue: TNotifyEvent);

    procedure SetOnMouseLeave(AValue: TNotifyEvent);

    procedure SetButtonSpace(AValue: Integer);

    procedure SetAutoArrangeButtons(AValue: Boolean);

    function GetHintColor: TColor;
    procedure SetHintColor(AValue: TColor);

    procedure SetSaturdayColor(AValue: TColor);
  public
    //berschriebene Routinen
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    //Neue Methoden

    //Fixe Feiertage (Verwaltung)
    procedure AddFixedHoliday(Day, Month: Byte); overload;
    procedure AddFixedHoliday(AHoliday: TDayMonthRecord); overload;
    procedure AddFixedHolidays(Holidays: Array of TDayMonthRecord);
    function FixedHolidayExists(Day, Month: Byte): Boolean; overload;
    function FixedHolidayExists(AHoliday: TDayMonthRecord): Boolean; overload;
    function AllDaysAreHolidays(AMonth: Byte): Boolean;
    procedure RemoveFixedHoliday(Day, Month: Byte); overload;
    procedure RemoveFixedHoliday(AHoliday: TDayMonthRecord); overload;
    procedure RemoveFixedHolidays(Holidays: Array of TDayMonthRecord);
    procedure RemoveAllFixedHolidays;

    //Dynamische Feiertage (Verwaltung)
    procedure AddDynamicHoliday(ADate: TDateTime);
    procedure AddDynamicHolidays(Dates: Array of TDateTime);
    function FindDynamicHoliday(ADate: TDateTime): Integer;
    procedure RemoveDynamicHoliday(ADate: TDateTime);
    procedure RemoveDynamicHolidays(Dates: Array of TDateTime);
    procedure RemoveAllDynamicHolidays;

    procedure RemoveAllHolidays;

    //'Vor' und 'Zurck' (Buttons)
    procedure NextMonth;
    procedure PreviousMonth;

    //Weitere 'Vor'- und 'Zurck'-Routinen
    procedure NextYear;
    procedure PreviusYear;
    procedure NextWeek;
    procedure PreviousWeek;
    procedure NextDay;
    procedure PreviousDay;
    procedure JumpToToday;

    //Auswahl
    procedure ResetSelection;
    function GetSelStartAsDateTime: TDateTime;
    function GetSelEndAsDateTime: TDateTime;
    procedure SelectDay(ADay: ShortInt);
    procedure StopSelecting;
    procedure SelectRow(ARow: TRowIndex);

    //Performance
    function TestPerformance(ChangeMonth: Boolean = false): Extended;

    //Neu zeichnen (v.a. fr Repaint bei TWinControl, das protected gesetzt wird)
    procedure Redraw;

    //Bewegliche Feiertage (Berechnung)

    //Ostern und davon abhngige Feiertage
    function GetEasterSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Ostern
    function GetEasterMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
    function GetHolySaturday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Karsamstag
    function GetGoodFriday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Karfreitag
    function GetMaundyThursday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Grndonnerstag
    function GetPalmSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Palmsonntag
    function GetAshWednesday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Aschermittwoch
    function GetShroveTuesday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Fasching(s)dienstag (Fastnacht)
    function GetCarnivalMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Rosenmontag
    function GetWhitsunSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Pfingsten
    function GetWhitsunMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
    function GetAscensionDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Christi Himmelfahrt
    function GetCorpusChristi(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Fronleichnam

    //Mutter- und Vatertag
    function GetMothersDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Muttertag
    function GetFathersDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Vatertag

    //Sommer- und Winterzeit
    function GetSummerTimeChangingDate(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Sommerzeit-Umstellung
    function GetWinterTimeChangingDate(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Winterzeit-Umstellung

    //Advent und davon abhngige Feiertage
    function GetFourthAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //4. Advent
    function GetThirdAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //3. Advent
    function GetSecondAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //2. Advent
    function GetFirstAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //1. Advent

    //Diverse Feiertage
    function GetThanksgiving(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Erntedankfest
    function GetTotensonntag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Totensonntag
    function GetVolkstrauertag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Volkstrauertag
    function GetBussundBettag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Bu- und Bettag (Deutschland)
    function GetBussundBettagSwitzerland(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime; //Bu- und Bettag (Schweiz)

    //Diverse
    procedure SetAutoRowHeight;
  protected
    //berschriebene Routinen
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    {$IFDEF WINCONTROL}
      procedure Repaint; override;
    {$ELSE}
      procedure Paint; override;
    {$ENDIF}

    //Abgefangene Messages (wie MouseEnter und MouseLeave)
    procedure _MouseEnter(var AMsg: TMessage); message CM_MOUSEENTER;
    procedure _MouseLeave(var AMsg: TMessage); message CM_MOUSELEAVE;
    procedure _Size(var AMsg: TMessage); message WM_SIZE;

    //Flackern verhindern (WM_ERASEBKGND abfangen)
    procedure _EraseBg(var AMsg: TWMEraseBkgnd); message WM_ERASEBKGND;

    {$IFDEF WINCONTROL}
      //Bei Bedarf neu zeichnen
      procedure _Paint(var AMsg: TMessage); message WM_PAINT;
    {$ENDIF}
  published
    property Date: TDateTime read FDate write SetDate;
    property OnChangeDate: TNotifyEvent read FOnChangeDate write SetOnChangeDate;
    property OnChangeDateEx: TDateChangeEvent read FOnChangeDateEx write SetOnChangeDateEx;
    property Font: TFont read FFont write SetFont;
    property BorderColor: TColor read FBorderColor write SetBorderColor default clBlack;
    property BackgroundColor: TColor read FBackgroundColor write SetBackgroundColor default clWhite;
    property SundayColor: TColor read FSundayColor write SetSundayColor default clRed;
    property WeekColor: TColor read FWeekColor write SetWeekColor default clBlue;
    property OutOfMonthColor: TColor read FOutOfMonthColor write SetOutOfMonthColor default clGray;
    property SelectedDayBold: Boolean read FSelectedDayBold write SetSelectedDayBold default true;
    property HeadlineStyle: TFontStyles read FHeadlineStyle write SetHeadlineStyle default [fsBold];
    property TitleFont: TFont read FTitleFont write SetTitleFont;
    property SelStartDay: ShortInt read FSelStartDay write SetSelStartDay default NoSelStart;
    property SelEndDay: ShortInt read FSelEndDay write SetSelEndDay default NoSelEnd;
    property SelectionBackgroundColor: TColor read FSelectionBackgroundColor write SetSelectionBackgroundColor default clSkyBlue;
    property AutoHeight: Boolean read FAutoHeight write SetAutoHeight default true;
    property HolidayColor: TColor read FHolidayColor write SetHolidayColor default clRed;
    property UseDoubleBuffer: Boolean read FUseDoubleBuffer write SetUseDoubleBuffer default true;
    property OnSelectDate: TSelectDateEvent read FOnSelectDate write SetOnSelectDate;
    property ApplyOnSelect: Boolean read FApplyOnSelect write SetApplyOnSelect default true;
    property MonthButtonsVisible: Boolean read FMonthButtonsVisible write SetMonthButtonsVisible default true;
    property YearButtonsVisible: Boolean read FYearButtonsVisible write SetYearButtonsVisible default true;
    property ButtonSize: Integer read FButtonSize write SetButtonSize default 20;
    property LastMonthButtonHint: String read GetLastMonthButtonHint write SetLastMonthButtonHint;
    property NextMonthButtonHint: String read GetNextMonthButtonHint write SetNextMonthButtonHint;
    property LastYearButtonHint: String read GetLastYearButtonHint write SetLastYearButtonHint;
    property NextYearButtonHint: String read GetNextYearButtonHint write SetNextYearButtonHint;
    property HighlightDate: Boolean read FHighlightDate write SetHighlightDate default true;
    property HeadlineOffset: Integer read FHeadlineOffset write SetHeadlineOffset default 10;
    property RowHeight: Integer read FRowHeight write SetRowHeight default 16;
    property ReactOnMouseDown: Boolean read FReactOnMouseDown write SetReactOnMouseDown default true;
    property ReactOnMouseUp: Boolean read FReactOnMouseUp write SetReactOnMouseUp default false;
    property OnRightClickDate: TRightClickDateEvent read FOnRightClickDate write SetRightClickDate;
    property ForceSixthWeekDraw: Boolean read FForceSixthWeekDraw write SetForceSixthWeekDraw default true;
    property IgnoreHolidays: Boolean read FIgnoreHolidays write SetIgnoreHolidays default false;
    property BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth default 2;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write SetOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write SetOnMouseLeave;
    property ButtonSpace: Integer read FButtonSpace write SetButtonSpace default 3;
    property AutoArrangeButtons: Boolean read FAutoArrangeButtons write SetAutoArrangeButtons default true;
    property EnableHints: Boolean read FEnableHints write FEnableHints default true;
    property OnCustomDateHint: TCustomDateHintEvent read FOnCustomDateHint write FOnCustomDateHint;
    property HintColor: TColor read GetHintColor write SetHintColor default clInfoBk;
    property OnBeforeMouseMove: TMouseMoveEvent read FOnBeforeMouseMove write FOnBeforeMouseMove;
    property OnBeforeMouseDown: TMouseEvent read FOnBeforeMouseDown write FOnBeforeMouseDown;
    property OnBeforeMouseUp: TMouseEvent read FOnBeforeMouseUp write FOnBeforeMouseUp;
    property OnAfterMouseMove: TMouseMoveEvent read FOnAfterMouseMove write FOnAfterMouseMove;
    property OnAfterMouseDown: TMouseEvent read FOnAfterMouseDown write FOnAfterMouseDown;
    property OnAfterMouseUp: TMouseEvent read FOnAfterMouseUp write FOnAfterMouseUp;
    property SelectOnMouseMove: Boolean read FSelectOnMouseMove write FSelectOnMouseMove default true;
    property OnSelectionChange: TNotifyEvent read FOnSelectionChange write FOnSelectionChange;
    property SaturdayColor: TColor read FSaturdayColor write SetSaturdayColor default clBlack;
    property AllowWeekSelection: Boolean read FAllowWeekSelection write FAllowWeekSelection default true;
  end;

//Registrierung in der Komponentenpalette
procedure Register;

//Fixe Feiertage (Deklarationen)
const
  //Generell (FH = fixed holidays)
  FH_NEWYEAR: TDayMonthRecord = (Day: 1; Month: 1); //Neujahr
  FH_EPIPHANY: TDayMonthRecord = (Day: 6; Month: 1); //Hl. Drei Knige
  FH_CANDLEMAS: TDayMonthRecord = (Day: 2; Month: 2); //Mari Lichtmess
  FH_FEASTOFTHEASSUMPTION: TDayMonthRecord = (Day: 15; Month: 8); //Mari Himmelfahrt
  FH_ALLSAINTSDAY: TDayMonthRecord = (Day: 1; Month: 11); //Allerheiligen
  FH_ALLSOULSDAY: TDayMonthRecord = (Day: 2; Month: 11); //Allerseelen
  FH_CHRISTMAS: TDayMonthRecord = (Day: 24; Month: 12); //Weihnachten
  FH_CHRISTMASDAY: TDayMonthRecord = (Day: 25; Month: 12); //Erster Weihnachtsfeiertag (Christtag)
  FH_BOXINGDAY: TDayMonthRecord = (Day: 26; Month: 12); //Zweiter Weihnachtsfeiertag (Stefanitag)
  FH_STSILVESTER: TDayMonthRecord = (Day: 31; Month: 12); //Silvester

  //sterreich (FHA = fixed holidays austria)
  FHA_STAATSFEIERTAG: TDayMonthRecord = (Day: 1; Month: 5); //Staatsfeiertag
  FHA_NATIONALFEIERTAG: TDayMonthRecord = (Day: 26; Month: 10); //Nationalfeiertag
  FHA_MARIAEEMPFAENGNIS: TDayMonthRecord = (Day: 8; Month: 12); //Mari Empfngnis

  //Deutschland (FHG = fixed holidays germany)
  FHG_TAGDERARBEIT: TDayMonthRecord = (Day: 1; Month: 5); //Tag der Arbeit
  FHG_TAGDERDEUTSCHENEINHEIT: TDayMonthRecord = (Day: 3; Month: 10); //Tag der deutschen Einheit
  FHG_REFORMATIONSTAG: TDayMonthRecord = (Day: 31; Month: 10); //Reformationstag

//Sonstige Deklarationen
const
  //Texte fr Kennzeichunungstyp (TDateFlag)
  SpecialFlags: Array [TDateFlag] of String = ('Keine Kennzeichnung', 'Auerhalb des Monats', 'Feiertag');

//Ende Interfaceteil, Beginn Implementationteil
implementation

procedure TDSCalendar.SetDate(AValue: TDateTime);
var
  Change: Boolean; //Fr OnChangeDateEx (nderung unterbinden)
begin
  Change := true; //nderung zulassen
  if Assigned(FOnChangeDateEx) then
    FOnChangeDateEx(Self, FDate, AValue, Change);
  if not Change then begin //nderungen unterbinden
    FNoSelStartAllowed := true; //SelStart darf nicht gesetzt werden
    exit; //Nichts ndern
    end;

  //Fr SelectOnMouseMove
  //Wenn Monat gewechselt -> Auswahl zurcksetzen
  if (MonthOf(AValue) <> MonthOf(FDate)) and (FMouseIsDown) then begin
    FSelStartDay := NoSelStart;
    FSelEndDay := NoSelEnd;
    FMonthChanged := true; //Monatsnderung signalisieren; wird in OnMouseMove wieder zurckgesetzt

    if Assigned(FOnSelectionChange) then //OnSelectionChange auslsen
      FOnSelectionChange(self);

    end;

  FDate := AValue;
  if Assigned(FOnChangeDate) then
    FOnChangeDate(Self);
  Repaint;
end;

constructor TDSCalendar.Create(AOwner: TComponent);
var
  _Self: TDSCalendar;
begin
  inherited;
  _Self := Self;

  //Initialisieren
  FIsDestroying := true; //True bis Komponente vollstndig erzeugt
  Parent := AOwner as TWinControl;
  {$IFDEF WINCONTROL}
  FCanvas := TCanvas.Create;
  FCanvas.Handle := GetDC(Handle);
  {$ENDIF}

  //DoubleBuffer erzeugen
  DoubleBuffer := TBitmap.Create;

  //Datum und Farben
  FDate := EncodeDate(2005, 7, 31);
  FBackgroundColor := clWhite;
  FBorderColor := clBlack;
  FBorderWidth := 2;

  //Schriftart
  FFont := TFont.Create;
  FFont.Name := 'Verdana';
  FFont.Size := 10;
  FFont.OnChange := OnFontChange;

  //Farben und Stile
  FSundayColor := clRed;
  FWeekColor := clBlue;
  FOutOfMonthColor := clGray;
  FSelectedDayBold := true;
  FHeadlineStyle := [fsBold];

  //Titelschriftart
  FTitleFont := TFont.Create;
  FTitleFont.Name := 'Verdana';
  FTitleFont.Size := 10;
  FTitleFont.Style := [fsBold];
  FTitleFont.OnChange := OnTitleFontChange;

  //Diverse Optionen
  FSelStartDay := NoSelStart;
  FSelEndDay := NoSelEnd;
  FSelectionBackgroundColor := clSkyBlue;
  FAutoHeight := true;
  FHolidayColor := clRed;
  FUseDoubleBuffer := true;
  FApplyOnSelect := true;
  FMonthButtonsVisible := true;
  FYearButtonsVisible := true;
  FHighlightDate := true;
  FHeadlineOffset := 10;
  FRowHeight := 16;
  FReactOnMouseUp := false;
  FReactOnMouseDown := true;
  FForceSixthWeekDraw := true;
  FIgnoreHolidays := false;
  FButtonSpace := 3;
  FAutoArrangeButtons := true;
  FEnableHints := true;
  FSelectOnMouseMove := true;
  FMouseIsDown := false;
  FMonthChanged := false;
  FSaturdayColor := clBlack;
  FNoSelStartAllowed := false;
  FAllowWeekSelection := true;

  //Buttons 'Vor' und zurck'

  //Monat
  FLastMonthButton := TButton.Create(Self); //Letztes Monat
  with FLastMonthButton do begin
    Parent := _Self;
    ShowHint := true;
    Hint := 'Letztes Monat';
    Caption := '<';
    OnClick := FLastMonthClick;
    end;

  FNextMonthButton := TButton.Create(Self); //Nchstes Monat
  with FNextMonthButton do begin
    Parent := _Self;
    ShowHint := true;
    Hint := 'Nchstes Monat';
    Caption := '>';
    OnClick := FNextMonthClick;
    end;

  //Jahr
  FLastYearButton := TButton.Create(Self); //Letztes Jahr
  with FLastYearButton do begin
    Parent := _Self;
    ShowHint := true;
    Hint := 'Letztes Jahr';
    Caption := '<<';
    OnClick := FLastYearClick;
    end;

  FNextYearButton := TButton.Create(Self); //Nchstes Jahr
  with FNextYearButton do begin
    Parent := _Self;
    ShowHint := true;
    Hint := 'Nchstes Jahr';
    Caption := '>>';
    OnClick := FNextYearClick;
    end;

  //Hintfenster
  FHintWnd := THintWindow.Create(Self);
  with FHintWnd do
    Color := clInfoBk;

  //Dynamische Feiertagsliste
  FHolidayList := TList.Create;

  FIsDestroying := false; //Komponente vollstndig erzeugt

  //Letzte Einstellungen (Buttongre und Ausrichtung)
  Width := 210;
  SetButtonSize(20);
  SetAutoHeight(true); //Automtische Grenberechnung
end;

destructor TDSCalendar.Destroy;
begin
  //Kein Zeichnen mehr
  FIsDestroying := true;

  FFont.Destroy;
  FTitleFont.Destroy;
  FNextMonthButton.Destroy;
  FLastMonthButton.Destroy;
  FNextYearButton.Destroy;
  FLastYearButton.Destroy;
  FHintWnd.Destroy;
  RemoveAllDynamicHolidays; //Speicher freigeben
  FHolidayList.Destroy;

  //DoubleBuffer freigeben
  DoubleBuffer.Free;
  {$IFDEF WINCONTROL}
    FCanvas.Free;
  {$ENDIF}
  inherited;
end;

{$IFDEF WINCONTROL}
  procedure TDSCalendar.Repaint;
{$ELSE}
  procedure TDSCalendar.Paint;
{$ENDIF}
var
  d, m, y: Word; //Tag, Monat und Jahr von FDate
  FirstDay: Word; //Wochentag des Monatsersten
  CurrentDay: ShortInt; //Aktueller Tag (fr Schleife)
  i, j: Integer; //Schleifenvariablen
  w: Integer; //Weite (BorderWidth mit bercksichtigt)
  MonthOver: Boolean; //Monat bereits vorber?
  MonthNotStarted: Boolean; //Monat noch nicht begonnen?
  WeekNr: Word; //Tempvariable fr Wochennr.
  Canv: TCanvas; //Canvas, auf das gezeichnet wird
  MaxDay: Integer; //Maximale Anzahl Tage, die gezeichnet wird
  CanvasHandle, DoubleBufferCanvas: THandle; //Handle des Canvas, auf dem gezeichnet wird (bentigt wegen Bug in TCanvas.GetHandle -> Speicherleck)
  TempIsHoliday: Boolean; //Tempvariable falls zu zeichnender Tag ein Feiertag ist
  SelStartEx, SelEndEx: ShortInt; //Auswahlbeginn und -ende; damit auch zB Start 20 und End 10 den Bereich 10-20 markiert
  MondayCounter: Byte; //Anzahl Montage (fr Montagsarray)
  WeekCounter: Byte; //Anzahl Wochennummern (fr Wochennummernarray)

  //DrawColumnItem
  //  Zeichnet einen Text an eine bestimmte Position der Zeichenflche.
  //  Dabei existieren 8 Spalten und beliebig viele Reihen (meist 6 + 1).
  //  Der Text wird automatisch in den Spalten zentriert, Column beginnt
  //  mit der Zhlung bei 0, Row bei 1 (einfacher fr die Schleife unten).
  procedure DrawColumnItem(Column, Row: Integer; Number: Integer; IsHoliday: Boolean = false; AlternativeText: String = ''; DontAddToCoords: Boolean = false);
  var
    Text: String;
    OldNumber: Integer; //Zum Speichern von Number
    _x, _y: Integer; //Ausgabe-Koordinaten
    NewAbsoluteDate: TDateTime; //Neues (echtes) Datum wenn Datum vor Monatsbeginn liegt
    Flag: TDateFlag; //Spezielle Kennzeichnung (schon vor zeichnen zuweisen)
    _xrect, _xrect2: Integer; //X-Koordinaten des aktuellen Items (Left und Right) fr Auswahlrechteck
  begin
    //Initialisieren, um Compilerwarnings zu vermeiden
    NewAbsoluteDate := 0;
    Flag := dfNone;

    //Number speichern
    OldNumber := Number;

    //Item-Anfang und -Ende berechnen
    _xrect := FBorderWidth + Trunc((Column) * (w / 8)); //Item-Anfang berechnen (Spaltenbeginn (angepasst an Achtel-Teilung der Zeichenflche))
    _xrect2 := FBorderWidth + Trunc((Column + 1) * (w / 8)); //Item-Ende berechnen (Nchste Spalte)

    //Zahl in String umwandeln falls Alternativstext leer, ansonsten Alternativtext
    if AlternativeText = '' then begin

      //Wenn Monat bereits vorber -> Tage des neuen Monats -> daher DaysInAMonth abziehen
      if (MonthOver) and not (Column = 0) {gilt nicht fr Wochennr.} then
        Dec(Number, DaysInAMonth(y, m));

      //Wenn Monat noch nicht begonnen -> Tage des alten Monats -> daher DaysInAMonth vom Vormonat
      if (MonthNotStarted) and not (Column = 0) {gilt nicht fr Wochennr.} then begin
        if m = 1 then begin //Spezialfall: Januar (Vormonat = Dezember!)
          Inc(Number, DaysInAMonth(y - 1, 12) - (FirstDay - 1));
          NewAbsoluteDate := EncodeDate(y - 1, 12, Number);
        end else begin //Normalfall: beliebiger anderer Monat
          Inc(Number, DaysInAMonth(y, m - 1) - (FirstDay - 1));
          NewAbsoluteDate := EncodeDate(y, m - 1, Number);
          end;
        end;

      //Monatstag konvertieren
      Text := IntToStr(Number)
    end else
      Text := AlternativeText;

    //Ist Tag ein Montag?
    if (Column = 1) and (Row > 0) and (Row < 7) then begin //Wenn ja zu Montagsarray hinzufgen
      Inc(MondayCounter); //Ein Montag mehr

      FMondaysDayArray[MondayCounter] := CurrentDay; //Tag speichern (SelStartDay-kompatibel)
      end;

    //Eigentliche Abme- und Zeichenroutine
    with Canv do begin

      //Position bestimmen
      _x := FBorderwidth + //Komponenten-Rand
            Trunc((Column + 1) * (w / 8)) - //Spalte (angepasst an Achtel-Teilung der Zeichenflche)
            Trunc((w / 8) / 2) - //Hlfte der Spalte (zwecks Zentrieren)
            TextWidth(Text) div 2; //Hlfte der Textbereite (ebenfalls zwecks Zentrieren)
      _y := FBorderWidth + //Komponenten-Rand
            FTitleFont.Size + FHeadlineOffset + //Schriftgre des Titels plus Abstand
            FRowHeight * Row; //Offset und Zeile (normalerweise 15 Pixel pro Zeile); wird fr Offset 0 bergeben: direktes Zeichnen in die Header der Spalten#

      //Wochennummernkoordinaten speichern
      if (Column = 0) and (Row > 0) and (Row < 7) then begin //Erste Spalte, ab erster Zeile
        Inc(WeekCounter); //Eine Wochennummer mehr

        //Koordinaten speichern
        FWeekCoords[WeekCounter].Left := _xrect + 1;
        FWeekCoords[WeekCounter].Top := _y + 1;
        FWeekCoords[WeekCounter].Right := _xrect2 - 1;
        FWeekCoords[WeekCounter].Bottom := _y + FRowHeight - 1;
        end;

      //Selection malen falls notwendig
      if (((CurrentDay >= SelStartEx) and (CurrentDay <= SelEndEx)) or //Tag zwischen SelStart und SelEnd
         ((CurrentDay = d) and (FHighlightDate))) and  //Aktuelles Datum markieren (falls eingestellt)
         ((Column <> 0) and (Row <> 0)) then begin //Um Markierung von Wochennr. und Wochentagen zu vermeiden

        //Auswahl zeichnen
        Brush.Color := FSelectionBackgroundColor;
        Pen.Color := Brush.Color; //Stiftfarbe = Hintergrundfarbe
        Rectangle(_xrect, _y, _xrect2, _y + FRowHeight) //Auswahlhintergrund zeichnen
      end else
        Brush.Color := FBackgroundColor;

      //Schriftfarbe ndern falls Monat noch nicht begonnen oder bereits zu Ende
      if ((MonthOver) or (MonthNotStarted)) and not (Column = 0) {gilt nicht fr Wochennr.} then begin
        Font.Color := FOutOfMonthColor; //Graue Zahlen um nchsten Monat zu signalisieren
        Flag := dfOutOfMonth;
        end;

      //Schriftfarbe ndern falls Feiertag
      if IsHoliday then begin
        Font.Color := FHolidayColor;
        Flag := dfHoliday;
        end;

      //Item zeichnen
      TextOut(_x, _y, Text);
      end;

    //Position des Items speichern fr OnDateSelect
    if (AlternativeText = '') and not (DontAddToCoords) then begin
      Inc(CoordCount);

      //Werte speichern (fr Maus-Events)
      with FDayCoords[CoordCount] do begin
        Day := CurrentDay;
        SpecialFlag := Flag;
        if not MonthNotStarted then
          NewAbsoluteDate := StartOfTheMonth(FDate) + OldNumber - 1; //Monatsbeginn + Anzahl Tage (Originalparameter) - 1 [da Anzahl]
        Date := NewAbsoluteDate;

        //Koordinaten: immer einen Pixel weniger, um berlappungen mit den anliegenden Tagen zu vermeiden
        Coords.Left := _xrect + 1;
        Coords.Top := _y + 1;
        Coords.Right := _xrect2 - 1;
        Coords.Bottom := _y + RowHeight - 1;
        end;
      end;
  end;

begin
  //Wenn Komponente zerstrt wird nicht mehr zeichnen
  if FIsDestroying then
    exit;

  //Doublebuffer initialisieren falls notwendig
  if FUseDoubleBuffer then begin

    //Falls Komponente erzeugt wird noch nicht zeichnen, da sonst AV, weil DoubleBuffer noch nicht existiert
    if DoubleBuffer = nil then
      exit;

    //Idente Gre setzen
    DoubleBuffer.Width := Width;
    DoubleBuffer.Height := Height;
    end;

  //Auf DoubleBuffer-Canvas zeichnen falls entsprechend eingestellt
  if (FUseDoubleBuffer) and (DoubleBuffer <> nil) then
      Canv := DoubleBuffer.Canvas
  else begin
    {$IFDEF WINCONTROL}
      Canv := FCanvas;
    {$ELSE}
      Canv := Canvas;
    {$ENDIF}
    end;

  //Initialisierung
  CoordCount := 0;
  MondayCounter := 0;
  WeekCounter := 0;

  for i := 1 to 6 do //Array fr Montage rcksetzen
    FMondaysDayArray[i] := NoSelStart;

  for i := 1 to 6 do begin //Array fr Wochennummernkoordinaten rcksetzen
    FWeekCoords[i].Left := 0;
    FWeekCoords[i].Top := 0;
    FWeekCoords[i].Right := 0;
    FWeekCoords[i].Bottom := 0;
    end;

  //SelStart und SelEnd intern tauschen falls notwendig (nur, wenn keiner der beiden auf NoSel... gesetzt)
  if (FSelStartDay > FSelEndDay) and (FSelStartDay <> NoSelStart) and (FSelEndDay <> NoSelEnd) then begin
    SelStartEx := FSelEndDay;
    SelEndEx := FSelStartDay
  end else begin
    SelStartEx := FSelStartDay;
    SelEndEx := FSelEndDay;
    end;

  with Canv do begin

    //Flche fllen
    MonthNotStarted := false; //Um Komplikationen mit dem Zeichnen der Tagesnamen etc. zu vermeiden

    //Variablen initialisieren und gewhlte Eigenschaften (Farbe etc. setzen)
    MonthOver := false;
    Brush.Color := FBackgroundColor;
    Pen.Color := FBorderColor;
    Rectangle(0, 0, Width, Height);
    w := Width - 2 * FBorderWidth; //Rand abziehen
    Brush.Style := bsClear;
    Font := FFont;

    //Datum zerlegen
    DecodeDate(FDate, y, m, d);

    //Monatsname und Jahr zeichnen
    Font := FTitleFont;
    TextOut(Width div 2 - TextWidth(LongMonthNames[m] + ' ' + FloatToStr(y)) div 2, FBorderWidth, LongMonthNames[m] + ' ' + FloatToStr(y));

    //Einstellungen zurcksetzen
    Font := FFont;
    Font.Style := FHeadlineStyle;

    //Wochentage zeichnen
    for i := 0 to 7 do begin //Wochtentage + Woche (= 8)
      case i of
        0: begin //Woche zeichnen
             Font.Color := FWeekColor;
             DrawColumnItem(i, 0, 0, false, 'W'); //'W' fr Woche
             continue; //Schleife erst bei nchstem Wert forsetzen, da ansonsten 'So' darbergemalt wird
           end;
        6: begin
             Font.Color := FSaturdayColor;
             DrawColumnItem(i, 0, 0, false, ShortDayNames[7]); //Samstag
           end;
        7: begin //Sonderfall: Sonntag (z.B. rot zeichnen)
             Font.Color := FSundayColor;
             DrawColumnItem(i, 0, 0, false, ShortDayNames[1]); //Sonntag
           end;
        else begin //Standardfall: Wochentagsnamen zeichnen
               Font.Color := FFont.Color;
               DrawColumnItem(i, 0, 0, false, ShortDayNames[i + 1]); //Wochentag (kurz)
             end;
        end;
      end;

    //Tage zeichnen

    //Notwendige Einstellungen (Variablen setzen)
    Font := FFont;
    FirstDay := DayOfTheWeek(EncodeDate(y, m, 1)); //Wochentag des Monatsersten bestimmen
    CurrentDay := -FirstDay + 1; //Tage im Vormonat bercksichtigen (+1 das der letzte des Vormonats SelStart 0 entspricht)
    j := 0;
    Brush.Color := FBackgroundColor;
    Brush.Style := bsSolid;

    //Endwert (zeichnen) bestimmen (Anzahl Tage im Monat)
    MaxDay := DaysInAMonth(y, m);

    //Werte entsprechend erhhen falls 6 Wochen erzwungen
    if FForceSixthWeekDraw then begin
      Inc(MaxDay, 7); //um 7 mehr damit 6.te Woche gezeichnet wird
      if MaxDay = 28 + 7 {bereits in voriger Zeile um 7 erhht} then
        Inc(MaxDay, 7); //Spezialfall Februar (z.B. 2010) - Februar endet am Sonntag -> nur vier Wochen -> MaxDay noch einmal erhhen
      end;

    //Eigentliches Zeichnen
    while (CurrentDay < MaxDay) and (j <= 5) {maximal 6 Wochen} do begin
      Inc(j); //Nchste Spalte

      //Woche zeichnen

      //Einstellungen (Variablen) setzen
      Font := FFont;
      Font.Color := FWeekColor;
      Brush.Color := FBackgroundColor;

      //Wochennr. bestimmen
      WeekNr := WeekOfTheYear(StartOfTheMonth(EncodeDate(y, m, d)) + CurrentDay + 1);

      //Wochennr. zeichnen
      DrawColumnItem(0, j, WeekNr, false, '', true); //Nicht zu Tageskoordinaten mit aufnehmen (sinnlos)

      if j <> 1 then //Bei Montag starten ab zweiter Zeile; in erster bei Monatserstem
        FirstDay := 1;

      //Tage vor Monatserstem (Vormonat) zeichnen
      MonthNotStarted := true;
      for i := 1 to FirstDay - 1 do begin
        Inc(CurrentDay); //Nchster Tag
        DrawColumnItem(i, j, i); //Zeichnen
        end;
      MonthNotStarted := false;

      //Tage zeichnen (zum Monat gehrige)

      for i := FirstDay {anderer Wert falls Monatserster nicht Montag} to 7 do begin

        //Variablen (Einstellungen) setzen
        TempIsHoliday := false; //Rcksetzen
        Inc(CurrentDay); //Nchster Tag
        Font := FFont; //Schriftart rcksetzen

        //Als Feiertag bergeben falls notwendig
        if not FIgnoreHolidays then
          if (CurrentDay in FFixedHolidays[m] {fixer Feiertag}) or (FindDynamicHoliday(EncodeDate(y, m, 1) + CurrentDay - 1) <> -1 {dynamischer Feiertag}) then
            TempIsHoliday := true;

        //Prfen, ob Tag auerhalb des angezeigten Monats liegt
        if CurrentDay > Abs(DaysInAMonth(y, m)) then //Abs um Compiler-Warnung zu vermeiden
          MonthOver := true; //Monat vorber -> Rest grau zeichnen

        //Markierung (falls eingestellt) da Tag von FDate
        if (CurrentDay = d) and FSelectedDayBold and not (MonthOver) then //Fett, wenn Tag von FDate
          Font.Style := Font.Style + [fsBold];

        //Tage zeichnen
        if i = 7 then begin //Sonderfall: Sonntag (z.B. rot zeichnen)
          if not MonthOver then //Nur dann z.B. rot zeichnen, wenn im aktuellen Monat, ansonsten grau
            Font.Color := FSundayColor;
          end;

        if i = 6 then begin //Sonderfall: Samstag (z.B. rot zeichnen)
          if not MonthOver then //Nur dann z.B. rot zeichnen, wenn im aktuellen Monat, ansonsten grau
            Font.Color := FSaturdayColor;
          end;

        DrawColumnItem(i, j, CurrentDay, TempIsHoliday) //Eigentliches Zeichnen
        end;
      end;
    end;

  //Inhalt von DoubleBuffer-Zeichenflche in Komponentencanvas kopieren
  if (FUseDoubleBuffer) and (DoubleBuffer <> nil) then begin
    CanvasHandle := GetDC(Handle);
    DoubleBufferCanvas := DoubleBuffer.Canvas.Handle;
    BitBlt(CanvasHandle, 0, 0, Width, Height, DoubleBufferCanvas, 0, 0, SRCCOPY);
    ReleaseDC(Handle, CanvasHandle);
    ReleaseDC(DoubleBuffer.Canvas.Handle, DoubleBufferCanvas);
    end;

  //Buttons neu zeichnen (nur bei TWinControl notwendig)
  {$IFDEF WINCONTROL}
    FLastMonthButton.Repaint;
    FNextMonthButton.Repaint;
    FLastYearButton.Repaint;
    FNextYearButton.Repaint;
  {$ENDIF}
end;

procedure TDSCalendar.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  Day: ShortInt; //Suchergebnis: Tag
  Date: TDateTime; //Suchergebnis: Datum
  Flag: TDateFlag; //Suchergebnis: Flag
begin
  inherited;

  FMouseIsDown := false; //Maustaste nicht mehr gedrckt

  if Assigned(FOnBeforeMouseUp) then
    FOnBeforeMouseUp(Self, Button, Shift, X, Y);

  if not FReactOnMouseUp then
    exit;

  //Suchwerte vorinitialisieren
  Day := NoSelStart; //= Ungltig
  Date := 0;
  Flag := dfNone;

  //Alle mglichen Tage prfen
  for i := Low(FDayCoords) to High(FDayCoords) do begin

    //Prfen, ob Koordinaten in diesem Tages-Rechtecke liegen
    if PtInRect(FDayCoords[i].Coords, Point(X, Y)) then begin

      //Suchergebnisse speichern
      Day := FDayCoords[i].Day;
      Date := FDayCoords[i].Date;
      Flag := FDayCoords[i].SpecialFlag;
      end;
    end;
  if (Button = mbLeft) and (Date <> 0) then begin

    //Falls Datum existiert und Property entsprechend gesetzt: Tag wechseln (Date-Property)
    if (FApplyOnSelect) then
      SetDate(Date);

    //Falls Datum existiert Ereignis OnDaySelect auslsen
    if Assigned(FOnSelectDate) then
      FOnSelectDate(Self, Day, X, Y, Date, Flag);
    end;

  //Falls Datum existiert Ereignis OnRightClickDate auslsen
  if (Button = mbRight) and (Date <> 0) and Assigned(FOnRightClickDate) then
    FOnRightClickDate(Self, X, Y, Date, Flag);

  if Assigned(FOnAfterMouseUp) then
    FOnAfterMouseUp(Self, Button, Shift, X, Y);
end;

procedure TDSCalendar.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  Day: ShortInt; //Suchergebnis: Tag
  Date: TDateTime; //Suchergebnis: Datum
  Flag: TDateFlag; //Suchergebnis: Flag
begin
  inherited;

  if Assigned(FOnBeforeMouseDown) then
    FOnBeforeMouseDown(Self, Button, Shift, X, Y);

  if not FReactOnMouseDown then
    exit;

  if Shift = [ssLeft] then
    FMouseIsDown := true; //Maustaste gedrckt

  //Suchwerte vorinitialisieren
  Day := NoSelStart; //= Ungltig
  Date := 0;
  Flag := dfNone;

  //Alle mglichen Tage prfen
  for i := Low(FDayCoords) to High(FDayCoords) do begin

    //Prfen, ob Koordinaten in diesem Tages-Rechtecke liegen
    if PtInRect(FDayCoords[i].Coords, Point(X, Y)) then begin

      //Suchergebnisse speichern
      Day := FDayCoords[i].Day;
      Date := FDayCoords[i].Date;
      Flag := FDayCoords[i].SpecialFlag;
      end;
    end;

  if (Button = mbLeft) and (Date <> 0) then begin

    //Falls Datum existiert Ereignis OnDaySelect auslsen
    if Assigned(FOnSelectDate) then
      FOnSelectDate(Self, Day, X, Y, Date, Flag);

    //Falls Datum existiert und Property entsprechend gesetzt: Tag wechseln (Date-Property)
    if FApplyOnSelect then
      SetDate(Date); //Datum setzen

    //Abbrechen, wenn Setzen von SelStartDay nicht erlaubt
    if FNoSelStartAllowed then begin
      FNoSelStartAllowed := false; //Rcksetzen
      exit; //Abbrechen
      end;

    if (FSelectOnMouseMove) and (FMouseIsDown) and (Day >= -7) and (Day <= 42) then begin //Auswahl mit Maus beginnen

      if FMonthChanged then begin //Wenn einen Tag im nchsten Monat markiert Auswahl rcksetzen
        FMonthChanged := false;
       end else begin //Markierung setzen
        SetSelStartDay(Day); //Markierungsanfang = markierter Tag
        SetSelEndDay(Day); //Auswahlanfang = Auswahlende (bei Markierungsbeginn)
        end;
      end;

    end;

  //Wenn Shifttaste gedrckt: altes Datum = SelStart; neues Datum = SelEnd
  if (FSelectOnMouseMove) and not (FMouseIsDown) and (Day >= -7) and (Day <= 42) then //Auswahl mit Maus beginnen
    SetSelEndDay(Day);

  //Wochenauswahl falls auf Woche geklickt
  if (Date = 0) and (FAllowWeekSelection) then begin //Kein Tag angeklickt

    //Alle mglichen Wochennummern prfen
    for i := Low(FWeekCoords) to High(FWeekCoords) do begin

      //Prfen, ob Koordinaten in diesem Tages-Rechtecke liegen
      if PtInRect(FWeekCoords[i], Point(X, Y)) then
        SelectRow(i); //Woche auswhlen
      end;

    end;

  //Falls Datum existiert Ereignis OnRightClickDate auslsen
  if (Button = mbRight) and (Date <> 0) and Assigned(FOnRightClickDate) then
    FOnRightClickDate(Self, X, Y, Date, Flag);

  if Assigned(FOnAfterMouseDown) then
    FOnAfterMouseDown(Self, Button, Shift, X, Y);
end;

procedure TDSCalendar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  Day: ShortInt; //Suchergebnis: Tag
  Date: TDateTime; //Suchergebnis: Datum
  Flag: TDateFlag; //Suchergebnis: Flag
  AHint: String; //Benutzerhint
  ARect: TRect; //Temp. Rechteck (fr Hints)
begin
  inherited;

  if Assigned(FOnBeforeMouseMove) then
    FOnBeforeMouseMove(Self, Shift, X, Y);

  if not FEnableHints then
    exit;

  //Suchwerte vorinitialisieren
  Day := NoSelStart; //= Ungltig
  Date := 0;
  Flag := dfNone;

  //Alle mglichen Tage prfen
  for i := Low(FDayCoords) to High(FDayCoords) do begin

    //Prfen, ob Koordinaten in diesem Tages-Rechtecke liegen
    if PtInRect(FDayCoords[i].Coords, Point(X, Y)) then begin

      //Suchergebnisse speichern
      Day := FDayCoords[i].Day;
      Date := FDayCoords[i].Date;
      Flag := FDayCoords[i].SpecialFlag;
      end;
    end;

  if (Shift = [] {nur MouseMove, kein Drag&Drop o..}) and (Date <> 0) then begin

    //Falls Datum existiert Ereignis OnCustomDateHint auslsen
    if Assigned(FOnCustomDateHint) then begin
      FOnCustomDateHint(Self, Day, X, Y, Date, Flag, AHint);

      //Hint anzeigen wenn nicht leer
      if AHint <> '' then begin

        //Hint anzeigen (10 Abstand wegen Mauszeiger)
        {$IFDEF WINCONTROL}
          with FCanvas do begin
        {$ELSE}
          with Canvas do begin
        {$ENDIF}
            ARect := FHintWnd.CalcHintRect(GetSystemMetrics(SM_CXSCREEN), AHint, nil); //Volle Bildschirmgre steht zur Verfgung

            //Koordinaten von Komponentenbereich auf Bildschirm umrechnen
            ARect.TopLeft := Self.ClientToScreen(ARect.TopLeft);
            ARect.BottomRight := Self.ClientToScreen(ARect.BottomRight);
            OffsetRect(ARect, X + 10, Y); //Um 10px verschieben damit Mauszeiger Text nicht verdeckt
            FHintWnd.ActivateHint(ARect, AHint); //Hintfenster aktivieren
            FHintWnd.Caption := AHint; //Hint zuweisen
          end;
        end;
      end;
  end else begin
    if (FSelectOnMouseMove) and (FMouseIsDown) and (Day >= -7) and (Day <= 42) then //Auswahl mit Mausbewegung ndern
      SetSelEndDay(Day); //Auswahlende auf Day (unter Cursor) setzen

    //Kein Hintfenster mehr anzeigen
    if FHintWnd.Caption <> '' then
      FHintWnd.ReleaseHandle;
    end;

  if Assigned(FOnAfterMouseMove) then
    FOnAfterMouseMove(Self, Shift, X, Y);
end;

procedure TDSCalendar._MouseEnter(var AMsg: TMessage);
begin
  if Assigned(FOnMouseEnter) then
    FOnMouseEnter(Self);
end;

procedure TDSCalendar._MouseLeave(var AMsg: TMessage);
begin
  if Assigned(FOnMouseLeave) then
    FOnMouseLeave(Self);
end;

procedure TDSCalendar._Size(var AMsg: TMessage);
begin
  if FIsDestroying then //Nichts tun, wenn erzeugt oder zerstrt wird
    exit;

  //Hhe rcksetzen falls AutoHeight eingestellt
  if FAutoHeight then begin
    SetAutoHeight(true);
    SetButtonSize(FButtonSize);
    end;
end;

{$IFDEF WINCONTROL}
  procedure TDSCalendar._Paint(var AMsg: TMessage);
  begin
    //Neuzeichnen (TControl zeichnet sich nicht selbst neu)
    Repaint;
  end;
{$ENDIF}

procedure TDSCalendar.OnFontChange(Sender: TObject);
begin
  SetAutoHeight(FAutoHeight); //Anpassen falls Schriftgre gendert
  Repaint;
end;

procedure TDSCalendar.OnTitleFontChange(Sender: TObject);
begin
  Repaint;
end;

procedure TDSCalendar.SetOnChangeDate(AValue: TNotifyEvent);
begin
  FOnChangeDate := AValue;
end;

procedure TDSCalendar.SetOnChangeDateEx(AValue: TDateChangeEvent);
begin
  FOnChangeDateEx := AValue;
end;

procedure TDSCalendar.SetFont(AValue: TFont);
begin
  FFont.Assign(AValue);
  SetAutoHeight(FAutoHeight); //Gre anpassen falls notwendig
  Repaint;
end;

procedure TDSCalendar.SetBorderColor(AValue: TColor);
begin
  if FBorderColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FBorderColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetBackgroundColor(AValue: TColor);
begin
  if FBackgroundColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FBackgroundColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetSundayColor(AValue: TColor);
begin
  if FSundayColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FSundayColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetWeekColor(AValue: TColor);
begin
  if FWeekColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FWeekColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetOutOfMonthColor(AValue: TColor);
begin
  if FOutOfMonthColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FOutOfMonthColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetSelectedDayBold(AValue: Boolean);
begin
  if FSelectedDayBold <> AValue then begin //Nur neuzeichnen wenn gendert
    FSelectedDayBold := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetHeadlineStyle(AValue: TFontStyles);
begin
  if FHeadlineStyle <> AValue then begin //Nur neuzeichnen wenn gendert
    FHeadlineStyle := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetTitleFont(AValue: TFont);
begin
  FTitleFont.Assign(AValue);
  SetAutoHeight(FAutoHeight); //Gre anpassen falls notwendig
  Repaint;
end;

procedure TDSCalendar.SetSelStartDay(AValue: ShortInt);
begin
  if FSelStartDay <> AValue then begin //Nur neuzeichnen wenn gendert
    FSelStartDay := AValue;

    if Assigned(FOnSelectionChange) then //OnSelectionChange auslsen
      FOnSelectionChange(Self);

    Repaint;
    end;
end;

procedure TDSCalendar.SetSelEndDay(AValue: ShortInt);
begin
  if FSelEndDay <> AValue then begin //Nur neuzeichnen wenn gendert
    FSelEndDay := AValue;

    if Assigned(FOnSelectionChange) then //OnSelectionChange auslsen
      FOnSelectionChange(Self);

    Repaint;
    end;
end;

procedure TDSCalendar.SetSelectionBackgroundColor(AValue: TColor);
begin
  if FSelectionBackgroundColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FSelectionBackgroundColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetAutoHeight(AValue: Boolean);
begin
  //Immer neuzeichen, da die Methode auch intern beim ndern von anderen Eigenschaften etc. verwendet wird
  FAutoHeight := AValue;
  if FAutoHeight then
    Height := 2 * FBorderWidth + //Rand
              FTitleFont.Size + //Titelschriftgre
              FHeadlineOffset + //Abstand Spaltenberschriften vom Titel
              7 * FRowHeight; //Zeilenhhe (Headings +  max. 6 Wochen)
  Repaint;
end;

procedure TDSCalendar.SetHolidayColor(AValue: TColor);
begin
  if FHolidayColor <> AValue then begin //Nur neuzeichnen wenn gendert
    FHolidayColor := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetUseDoubleBuffer(AValue: Boolean);
begin
  if FUseDoubleBuffer <> AValue then //Nur setzen wenn gendert
    FUseDoubleBuffer := AValue;
end;

procedure TDSCalendar.SetOnSelectDate(AValue: TSelectDateEvent);
begin
  FOnSelectDate := AValue;
end;

procedure TDSCalendar.SetApplyOnSelect(AValue: Boolean);
begin
  FApplyOnSelect := AValue;
  Repaint; //Neuzeichnen (betrifft andere Eigenschaften)
end;

procedure TDSCalendar.SetMonthButtonsVisible(AValue: Boolean);
begin
  FMonthButtonsVisible := AValue;
  FNextMonthButton.Visible := FMonthButtonsVisible;
  FLastMonthButton.Visible := FMonthButtonsVisible;

  //Automatisch neu anordnen falls eingestellt
  SetAutoArrangeButtons(FAutoArrangeButtons);

  //Neuzeichnen muss nicht mehr aufgerufen werden da bereits von SetAutoArrangeButtons aufgerufen
end;

procedure TDSCalendar.SetYearButtonsVisible(AValue: Boolean);
begin
  FYearButtonsVisible := AValue;
  FNextYearButton.Visible := FYearButtonsVisible;
  FLastYearButton.Visible := FYearButtonsVisible;
end;

procedure TDSCalendar.SetButtonSize(AValue: Integer);
begin
  FButtonSize := AValue;

  //Monate
  FLastMonthButton.Left := FBorderWidth;
  FLastMonthButton.Top := FBorderWidth;
  FLastMonthButton.Width := FButtonSize;
  FLastMonthButton.Height := FButtonSize;

  FNextMonthButton.Left := Width - (FBorderWidth + FButtonSize);
  FNextMonthButton.Top := FBorderWidth;
  FNextMonthButton.Width := FButtonSize;
  FNextMonthButton.Height := FButtonSize;

  //Jahre
  FLastYearButton.Left := FBorderWidth + FButtonSize + FButtonSpace;
  FLastYearButton.Top := FBorderWidth;
  FLastYearButton.Width := FButtonSize;
  FLastYearButton.Height := FButtonSize;

  FNextYearButton.Left := Width - (FBorderWidth + FButtonSize) - (FButtonSize + FButtonSpace);
  FNextYearButton.Top := FBorderWidth;
  FNextYearButton.Width := FButtonSize;
  FNextYearButton.Height := FButtonSize;

  //Jahresbuttons neu anordnen, falls Monatsbuttons unsichtbar
  if (FAutoArrangeButtons) and (not FMonthButtonsVisible) then begin

    //Left der Jahresbuttons anpassen
    FLastYearButton.Left := FBorderWidth;
    FNextYearButton.Left := Width - (FBorderWidth + FButtonSize);
    end;
end;

function TDSCalendar.GetLastMonthButtonHint: String;
begin
  Result := FLastMonthButton.Hint;
end;

procedure TDSCalendar.SetLastMonthButtonHint(AValue: String);
begin
  FLastMonthButton.Hint := AValue;
end;

function TDSCalendar.GetLastYearButtonHint: String;
begin
  Result := FLastYearButton.Hint;
end;

procedure TDSCalendar.SetLastYearButtonHint(AValue: String);
begin
  FLastYearButton.Hint := AValue;
end;

function TDSCalendar.GetNextMonthButtonHint: String;
begin
  Result := FNextMonthButton.Hint;
end;

procedure TDSCalendar.SetNextMonthButtonHint(AValue: String);
begin
  FNextMonthButton.
  Hint := AValue;
end;

function TDSCalendar.GetNextYearButtonHint: String;
begin
  Result := FNextYearButton.Hint;
end;

procedure TDSCalendar.SetNextYearButtonHint(AValue: String);
begin
  FNextYearButton.Hint := AValue;
end;

procedure TDSCalendar.SetHighlightDate(AValue: Boolean);
begin
  //Nur neuzeichnen wenn gendert
  if FHighlightDate <> AValue then begin
    FHighlightDate := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetHeadlineOffset(AValue: Integer);
begin
  FHeadlineOffset := AValue;
  SetAutoHeight(FAutoHeight); //Gre neu berechnen falls notwendig
end;

procedure TDSCalendar.SetRowHeight(AValue: Integer);
begin
  FRowHeight := AValue;
  SetAutoHeight(FAutoHeight); //Gre neu berechnen falls notwendig
end;

procedure TDSCalendar.SetReactOnMouseDown(AValue: Boolean);
begin
  FReactOnMouseDown := AValue;
end;

procedure TDSCalendar.SetReactOnMouseUp(AValue: Boolean);
begin
  FReactOnMouseUp := AValue;
end;

procedure TDSCalendar.SetRightClickDate(AValue: TRightClickDateEvent);
begin
  FOnRightClickDate := AValue;
end;

procedure TDSCalendar.SetForceSixthWeekDraw(AValue: Boolean);
begin
  //Nur neuzeichnen wenn gendert
  if FForceSixthWeekDraw <> AValue then begin
    FForceSixthWeekDraw := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetIgnoreHolidays(AValue: Boolean);
begin
  //Nur neuzeichnen wenn gendert
  if FIgnoreHolidays <> AValue then begin
    FIgnoreHolidays := AValue;
    Repaint;
    end;
end;

procedure TDSCalendar.SetBorderWidth(AValue: TBorderWidth);
begin
  //Nur neuzeichnen wenn gendert
  if FBorderWidth <> AValue then begin
    FBorderWidth := AValue;
    SetAutoHeight(FAutoHeight); //Gre neu berechnen falls notwendig
    SetButtonSize(FButtonSize); //Buttons neu anordnen falls notwendig
    end;
end;

procedure TDSCalendar.SetOnMouseEnter(AValue: TNotifyEvent);
begin
  FOnMouseEnter := AValue;
end;

procedure TDSCalendar.SetOnMouseLeave(AValue: TNotifyEvent);
begin
  FOnMouseLeave := AValue;
end;

procedure TDSCalendar.SetButtonSpace(AValue: Integer);
begin
  FButtonSpace := AValue;
  SetButtonSize(FButtonSize); //Buttongre neu setzen um nderung wirksam zu machen
end;

procedure TDSCalendar.SetAutoArrangeButtons(AValue: Boolean);
begin
  FAutoArrangeButtons := AValue;
  SetButtonSize(FButtonSize); //Buttongre neu setzen um nderung wirksam zu machen
end;

procedure TDSCalendar.SetSaturdayColor(AValue: TColor);
begin
  if AValue <> FSaturdayColor then begin //Nur neuzeichnen wenn gendert
    FSaturdayColor := AValue;
    Repaint; //Neuzeichnen
    end;
end;

function TDSCalendar.GetHintColor: TColor;
begin
  Result := FHintWnd.Color;
end;

procedure TDSCalendar.SetHintColor(AValue: TColor);
begin
  FHintWnd.Color := AValue;
end;

procedure TDSCalendar.FLastMonthClick(Sender: TObject);
begin
  SetDate(IncMonth(FDate, -1));
end;

procedure TDSCalendar.FNextMonthClick(Sender: TObject);
begin
  SetDate(IncMonth(FDate));
end;

procedure TDSCalendar.FLastYearClick(Sender: TObject);
begin
  SetDate(IncYear(FDate, -1));
end;

procedure TDSCalendar.FNextYearClick(Sender: TObject);
begin
  SetDate(IncYear(FDate));
end;

procedure TDSCalendar.AddFixedHoliday(Day, Month: Byte);
begin
  if (Month >= 1) and (Month <= 12) then //In Gltigkeitsbereich?
    FFixedHolidays[Month] := FFixedHolidays[Month] + [Day];
  Repaint;  
end;

procedure TDSCalendar.AddFixedHoliday(AHoliday: TDayMonthRecord);
begin
  AddFixedHoliday(AHoliday.Day, AHoliday.Month);
end;

procedure TDSCalendar.AddFixedHolidays(Holidays: Array of TDayMonthRecord);
var
  i: Integer;
begin
  for i := Low(Holidays) to High(Holidays) do
    AddFixedHoliday(Holidays[i].Day, Holidays[i].Month);
end;

function TDSCalendar.FixedHolidayExists(Day, Month: Byte): Boolean;
begin
  Result := Day in FFixedHolidays[Month];
end;

function TDSCalendar.FixedHolidayExists(AHoliday: TDayMonthRecord): Boolean;
begin
  Result := FixedHolidayExists(AHoliday.Day, AHoliday.Month);
end;

procedure TDSCalendar.RemoveFixedHoliday(Day, Month: Byte);
begin
  if (Month >= 1) and (Month <= 12) then //In Gltigkeitsbereich?
    FFixedHolidays[Month] := FFixedHolidays[Month] - [Day];
  Repaint;
end;

procedure TDSCalendar.RemoveFixedHoliday(AHoliday: TDayMonthRecord);
begin
  RemoveFixedHoliday(AHoliday.Day, AHoliday.Month);
end;

procedure TDSCalendar.RemoveFixedHolidays(Holidays: Array of TDayMonthRecord);
var
  i: Integer;
begin
  for i := Low(Holidays) to High(Holidays) do
    RemoveFixedHoliday(Holidays[i].Day, Holidays[i].Month);
  Repaint;
end;

procedure TDSCalendar.RemoveAllFixedHolidays;
var
  i: Byte;
begin
  for i := 1 to 12 do
    FFixedHolidays[i] := [];
  Repaint;
end;

procedure TDSCalendar.NextMonth;
begin
  FNextMonthClick(Self);
end;

procedure TDSCalendar.PreviousMonth;
begin
  FLastMonthClick(Self);
end;

procedure TDSCalendar.ResetSelection;
begin
  FSelStartDay := NoSelStart;
  FSelEndDay := NoSelEnd;

  if Assigned(FOnSelectionChange) then //OnSelectionChange auslsen
    FOnSelectionChange(self);

  Repaint;
end;

function TDSCalendar.TestPerformance(ChangeMonth: Boolean = false): Extended;
var
  Start, Ende, f: Int64;
  x: Integer;
begin
  QueryPerformanceCounter(Start); //Messung Start

  //1000x neu zeichnen bzw. Monat wechseln
  for x := 1 to 1000 do
    if ChangeMonth then
      NextMonth
    else
      Repaint;

  QueryPerformanceCounter(Ende); //Messung Ende
  QueryPerformanceFrequency(f); //Frequenz bestimmen

  Result := (Ende - Start) / f;
end;

procedure TDSCalendar.Redraw;
begin
  Repaint;
end;

procedure TDSCalendar.NextYear;
begin
  FNextYearClick(Self);
end;

procedure TDSCalendar.PreviusYear;
begin
  FLastYearClick(Self);
end;

procedure TDSCalendar.NextWeek;
begin
  SetDate(FDate + 7);
end;

procedure TDSCalendar.PreviousWeek;
begin
  SetDate(FDate - 7);
end;

procedure TDSCalendar.NextDay;
begin
  SetDate(FDate + 1);
end;

procedure TDSCalendar.PreviousDay;
begin
  SetDate(FDate - 1);
end;

function TDSCalendar.GetEasterSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
//Bemerkung: modifizierte Version vom Quelltext des Artikels "Wann ist Ostern" ('Der Entwickler' 1/99)
var
  gz, jh, x, z, ep, d, tag, monat: Integer;
begin
  gz := (AYear mod 19) + 1; //Goldene Zahl
  jh := (AYear div 100) + 1; //Jahrhundert
  x := ((jh * 3) div 4) - 12; //Ausgelassene Schaltjahre (um mit Sonne synchron zu bleiben)
  z := (((jh * 8) + 5) div 25) - 5; //Korrektur (Mondsynchronisation)
  ep := (((gz * 11) + 20 + z - x) mod 30); //Epact
  d := ((AYear * 5) div 4) - 10 - x; //Bestimmung eines Sonntags
  tag := 44 - ep; //Monatstag (Frhlingsbeginn)
  if tag < 21 then //Falls vor dem 21. Mrz (Frhlingsbeginn) 31 hinzuzhlen
    Inc(tag, 31);
  tag := tag + 7 - ((d + tag) mod 7); //Sonntag (Ostersonntag) ermitteln
  monat := 3; //Standardgem Monat Mrz
  if (tag >= 32) and (tag <= 61) then begin //Falls nach Mrz -> Aprildatum
    Dec(tag, 31); //Anzahl Tage fr Aprildatum abziehen
    monat := 4; //April
    end;
  Result := EncodeDate(AYear, monat, tag); //Ergebnis in Datum umwandeln

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetEasterMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) + 1;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetHolySaturday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 1;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetGoodFriday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 2;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetMaundyThursday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 3;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetPalmSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 7; //Sonntag der Vorwoche

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetAshWednesday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 46;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetShroveTuesday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 47;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetCarnivalMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) - 48;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetWhitsunSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) + 49;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetWhitsunMonday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetWhitsunSunday(AYear) + 1;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetAscensionDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) + 39;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetCorpusChristi(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetEasterSunday(AYear) + 60;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetMothersDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Muttertag (zweiten Sonntag im Mai) berechnen
  Result := EncodeDate(AYear, 5, 1); //1. Mai
  Result := EndOfTheWeek(Result); //Ende der Woche berechnen -> erster Sonntag
  Result := Result + 7; //7 Tage weiter -> zweiter Sonntag

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetFathersDay(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Vatertag (zweiten Sonntag im Juni) berechnen
  Result := EncodeDate(AYear, 6, 1); //1. Juni
  Result := EndOfTheWeek(Result); //Ende der Woche berechnen -> erster Sonntag
  Result := Result + 7; //7 Tage weiter -> zweiter Sonntag

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetSummerTimeChangingDate(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Sommerzeit-Umstellung (letzten Sonntag im Mrz) berechnen
  Result := EncodeDate(AYear, 3, 31); //Ausgehend vom 31. Mrz
  while DayOfTheWeek(Result) <> 7 {Sonntag} do
    Result := Result - 1; //Vorhergehenden Tag versuchen

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetWinterTimeChangingDate(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Sommerzeit-Umstellung (letzten Sonntag im Oktober) berechnen
  Result := EncodeDate(AYear, 10, 31); //Ausgehend vom 31. Oktober
  while DayOfTheWeek(Result) <> 7 {Sonntag} do
    Result := Result - 1; //Vorhergehenden Tag versuchen

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetFourthAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Vierten Advent (Sonntag vor Weihnachten) berechnen
  Result := EncodeDate(AYear, 12, 24); //Ausgehend von Weihnachten (24.12.)
  while DayOfTheWeek(Result) <> 7 {Sonntag} do
    Result := Result - 1; //Vorhergehenden Tag versuchen

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetThirdAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetFourthAdventSunday(AYear) - 7;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetSecondAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetFourthAdventSunday(AYear) - 14;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetFirstAdventSunday(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetFourthAdventSunday(AYear) - 21;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetThanksgiving(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Erntedankfest (erster Sonntag im September) berechnen
  Result := EncodeDate(AYear, 9, 1); //1. September
  Result := EndOfTheWeek(Result); //Ende der Woche berechnen -> erster Sonntag

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetTotensonntag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetFourthAdventSunday(AYear) - 28;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetVolkstrauertag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  Result := GetFourthAdventSunday(AYear) - 35;

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetBussundBettag(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Bu- und Bettag (Mittwoch zwischen 16. und 22.11.) berechnen
  Result := EncodeDate(AYear, 11, 16); //Ausgehend vom 16. November
  while DayOfTheWeek(Result) <> 3 {Mittwoch} do
    Result := Result  + 1; //Nchsten Tag versuchen

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

function TDSCalendar.GetBussundBettagSwitzerland(AYear: Word; AddToHolidays: Boolean = false; SwitchToHoliday: Boolean = false): TDateTime;
begin
  //Bu- und Bettag Schweiz (zweiter Sonntag im September) berechnen
  Result := EncodeDate(AYear, 9, 1); //1. September
  Result := EndOfTheWeek(Result); //Ende der Woche berechnen -> erster Sonntag
  Result := Result + 7; //7 Tage weiter -> zweiter Sonntag

  if AddToHolidays then //Datum automatisch zu den dynamischen Feiertagen hinzufgen
    AddDynamicHoliday(Result);

  if SwitchToHoliday then begin //Zu Datum springen
    ResetSelection; //Auswahl zurcksetzen
    SetDate(Result);
    end;
end;

procedure TDSCalendar.SetAutoRowHeight;
begin
  SetRowHeight(Round(FFont.Size * 1.5) + 1);
end;

procedure TDSCalendar.AddDynamicHoliday(ADate: TDateTime);
var
  FAPDate: PDateTime;
begin
  New(FAPDate); //Neuer Zeiger
  FAPDate^ := ADate;
  FHolidayList.Add(FAPDate); //Zur Liste hinzufgen
  Repaint; //Neu zeichnen
end;

procedure TDSCalendar.AddDynamicHolidays(Dates: Array of TDateTime);
var
  i: Integer;
begin
  for i := Low(Dates) to High(Dates) do
    AddDynamicHoliday(Dates[i]);
end;

function TDSCalendar.FindDynamicHoliday(ADate: TDateTime): Integer;
var
  i: Integer;
begin
  //Initialisieren (-1 = nichts gefunden)
  Result := -1;

  //Falls aufgerufen bevor Liste erstellt wurde Funktion verlassen um keine Fehler zu verursachen
  if FHolidayList = nil then
    exit;

  for i := FHolidayList.Count - 1 downto 0 do begin
    if TDateTime(FHolidayList[i]^) = ADate then begin //Datum gefunden
      Result := i; //Index zurckgeben
      exit; //Gefunden -> Funktion verlassen
      end;
    end;
end;

procedure TDSCalendar.RemoveDynamicHoliday(ADate: TDateTime);
var
  x: Integer; //temp.
  temp: PDateTime;
begin
  x := FindDynamicHoliday(ADate); //Datum suchen

  while x <> -1 {Datum gefunden} do begin
    temp := FHolidayList[x]; //Zeiger auslesen
    Dispose(temp); //Speicher freigeben
    FHolidayList.Delete(x); //Item aus Liste lschen
    x := FindDynamicHoliday(ADate); //Nchstes Vorkommnis suchen
    end;

  Repaint; //Neu zeichnen
end;

procedure TDSCalendar.RemoveDynamicHolidays(Dates: Array of TDateTime);
var
  i: Integer;
begin
  for i := Low(Dates) to High(Dates) do
    RemoveDynamicHoliday(Dates[i]);
end;

procedure TDSCalendar.RemoveAllDynamicHolidays;
var
  i: Integer;
  temp: PDateTime;
begin
  for i := FHolidayList.Count - 1 downto 0 do begin
    temp := FHolidayList[i]; //Zeiger auslesen
    Dispose(temp); //Speicher freigeben
    FHolidayList.Delete(i); //Item aus Liste lschen
    end;

  Repaint; //Neu zeichnen
end;

procedure TDSCalendar.RemoveAllHolidays;
begin
  //Dynamische und fixe Feiertage (mit den entsprechenden Routinen lschen)
  RemoveAllFixedHolidays;
  RemoveAllDynamicHolidays;
end;

function TDSCalendar.GetSelStartAsDateTime: TDateTime;
begin
  Result := StartOfTheMonth(FDate) - 1 + FSelStartDay;
end;

function TDSCalendar.GetSelEndAsDateTime: TDateTime;
begin
  Result := StartOfTheMonth(FDate) - 1 + FSelEndDay;
end;

procedure TDSCalendar.JumpToToday;
begin
  SetDate(SysUtils.Date);
end;

function TDSCalendar.AllDaysAreHolidays(AMonth: Byte): Boolean;
var
  DaysInM: Byte; //Tage im angegebenen Monat
  i: Byte;
begin
  DaysInM := DaysInMonth(AMonth);
  Result := true;
  for i := 1 to DaysInM do
    Result := Result and (i in FFixedHolidays[AMonth]); //Wenn einmal false auftritt kommt durch das "Ver-UNDen" false heraus
end;

procedure TDSCalendar.SelectDay(ADay: ShortInt);
begin //Auswahlstart = Auswahlende
  FSelStartDay := ADay;
  SetSelEndDay(ADay);
end;

procedure TDSCalendar.StopSelecting;
begin
  FMouseIsDown := false;
end;

procedure TDSCalendar._EraseBg(var AMsg: TWMEraseBkgnd);
begin
  if FIsDestroying then //Nichts tun, wenn erzeugt oder zerstrt wird
    exit;

  //Um Darstellungsfehler zu vermeiden Rand um Buttons neu zeichnen
  {$IFDEF WINCONTROL}
    with FCanvas do begin
  {$ELSE}
    with Canvas do begin
  {$ENDIF}

    Handle := AMsg.DC; //Canvas-Handle zuweisen (wird von AMsg bergeben)

    //Farben setzen
    Brush.Color := FBackgroundColor;
    Pen.Color := FBackgroundColor;

    //Buttonrnder neu zeichnen wenn sichtbar
    if FNextMonthButton.Visible then
      FillRect(FNextMonthButton.BoundsRect);
    if FNextYearButton.Visible then
      FillRect(FNextYearButton.BoundsRect);
    if FLastMonthButton.Visible then
      FillRect(FLastMonthButton.BoundsRect);
    if FLastYearButton.Visible then
      FillRect(FLastYearButton.BoundsRect);
    end;

  AMsg.Result := 1; //Um Flackern (Lschen des Hintergrunds) zu verhindern
end;

procedure TDSCalendar.SelectRow(ARow: TRowIndex);
begin
  FSelStartDay := FMondaysDayArray[ARow]; //Entsprechenden Montag auslesen (Auswahlbeginn)
  SetSelEndDay(FSelStartDay + 6); //Sonntag = Auswahlende (6 Tage spter)
end;

procedure Register;
begin
  //Komponente registrieren
  RegisterComponents('Dust Signs', [TDSCalendar]);
end;

end.
